/*
 * oscp-slave-fc14.c
 * Copyright (C) 2015 KotCzarny <tjosko@yahoo.com>
 *
 *  This program is free software; you can redistribute it and/or modify it
 *  under the terms of the GNU General Public License as published by the
 *  Free Software Foundation; either version 2 of the License, or (at your
 *  option) any later version.
 *
 *  This program is distributed in the hope that it will be useful, but
 *  WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 *  General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License along
 *  with this program; if not, write to the Free Software Foundation, Inc.,
 *  59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 *
 */

#include "oscp-slave.h"

#include <fc14audiodecoder.h>

void * fc14c;
#define fc14_dbuf_size 1024*2
char fc14_dbuf[fc14_dbuf_size];

int fc14state=0; float fc14cur=0, fc14dur=0; int fc14i=0;
int fc14song=0;

/////################
void fc14_show_supp() {
 printf("  [slave-fc14]: based on libfc14audiodecoder-1.0.3");
 printf(", formats: fc, fc13, fc14\n");
}

/////###########
//void fc14_flush() { lf("fc14_flush"); adev_flush(); }
/////###########
void fc14_init() { lf("fc14_init"); fc14state=0; dlog(9, "fc14_init"); if(fc14state>0) return;

 if(fc14c) dlog(9, "fc14_init: fc14c not null?");

 fc14c=fc14dec_new();

 // Initialize decoder's audio sample mixer.  frequency : output sample frequency
 // precision : bits per sample  channels : 1=mono, 2=stereo
 // zero : value of silent output sample (e.g. 0x80 for unsigned 8-bit, 0x0000 for signed 16-bit)
 fc14dec_mixer_init(fc14c, 44100, 16, 2, 0x0000);

 if(adev_init()<=0) { fc14state=0; return; }

 fc14state=1;
}
/////##########
void fc14_stop() { lf("fc14_stop");
 dlog(9, "fc14_stop");
 if(fc14state<=1) return;

 if(fc14state>=3) { 
  fc14dec_restart(fc14c);
 }

 fc14state=1;
}
/////#########
void fc14_end() { lf("fc14_end");
 dlog(9, "fc14_end"); if(fc14state>1) fc14_stop(); adev_end();

 // Delete decoder object.
 if(fc14c) fc14dec_delete(fc14c);

 fc14state=0; 
}
/////##################
void fc14_die(char *msg) { lf2("fc14_die",msg); dlog(9, "fc14_die"); fc14_stop(); cs(msg); }

char fc14fn[PATH_MAX]; int fc14fs=0;

/////##########################
void fc14_update_info(char *fn) { lf("fc14_update_info");
 static char lmsg[4096];

 fc14dur=fc14dec_duration(fc14c)/1000.0;
 sprintf(lmsg, "ANS_length=%.1f\n", fc14dur); cs(lmsg);

 snprintf(lmsg,1022,"Title: %s\n", get_basename(fn)); cs(lmsg);
 int fs=get_filesize(fn);
 snprintf(lmsg,1022,"Comment: %d, %s\n", fs, fc14dec_format_name(fc14c)); cs(lmsg);
}

void fc14_play();
/////################################
void fc14_load(char *input_filename0) { lfd("fc14_load:state",fc14state);
 static char lmsg[4096]; lmsg[4095]=0;
 snprintf(lmsg, 4095, "Playing %s\n", input_filename0); cs(lmsg);

 if(input_filename0[0]==0) return;
 if(input_filename0[0]==*"\"" && input_filename0[1]==*"\"" && input_filename0[2]==0)
 { cs("EOF code: stop requested\n"); if(fc14state>2) fc14_stop(); return; }

 if(fc14state==0) fc14_init();
 else if(fc14state>2) fc14_stop();

 fc14state=2; fc14cur=0; fc14dur=0;
 char *input_filename=input_filename0;
 int ret=0;


 fc14fs=get_filesize(input_filename);
 if(fc14fs<=0) { fc14_die("EOF code: invalid module size\n"); return; }

 char *module=malloc(fc14fs); if(!fc14fs) { fc14_die("EOF code: malloc failed\n"); return; }
 FILE *f=fopen(input_filename, "r"); if(!f) { fc14_die("EOF code: fopen failed\n"); return; }
 int ret0=fread(module, fc14fs, 1, f); fclose(f);

 // Apply input format header check to a memory buffer. Returns: 0 = recognized data, 1 = unknown data
 ret=fc14dec_detect(fc14c, module, fc14fs);

 // Initialize decoder with input data from a memory buffer. Input buffer may be freed as buffer
 // contents are copied. Returns: 0 = success, 1 = failure
 int ret2=0;
// if(ret==0)
 ret2=fc14dec_init(fc14c, module, fc14fs);

 free(module);
// if(ret!=0 || ret2!=0)
 if(ret==0 || ret2==0)
 {
sprintf(lmsg, "fc14 fs:%d, ret0:%d, ret:%d, ret2:%d", fc14fs, ret0, ret, ret2); dlog(9,lmsg);
 fc14_die("EOF code: fc14_load failed\n"); return; }

 sfmt=AV_SAMPLE_FMT_S16;
      if(sfmt==AV_SAMPLE_FMT_U8 || sfmt==AV_SAMPLE_FMT_U8P){ sformat.bits=8; }
 else if(sfmt==AV_SAMPLE_FMT_S16 || sfmt==AV_SAMPLE_FMT_S16P){ sformat.bits=16; }
 else if(sfmt==AV_SAMPLE_FMT_S32 || sfmt==AV_SAMPLE_FMT_S32P){ sformat.bits=32; }
 else if(sfmt==AV_SAMPLE_FMT_FLT || sfmt==AV_SAMPLE_FMT_FLTP){ sformat.bits=32; }
 else if(sfmt==AV_SAMPLE_FMT_DBL || sfmt==AV_SAMPLE_FMT_DBLP){ sformat.bits=64; }
 else { fc14_die("EOF code: Unsupported format\n"); return; }

 fc14cur=-1;
// fc14dur=fc14Info_GetDuration(fc14I, fc14Info_GetDefaultSong(fc14I))/1000.0;

 sformat.channels=2; sformat.rate=44100;
#ifdef LAO
 sformat.byte_format=AO_FMT_NATIVE; sformat.matrix=0;
#endif

 adev_setp(sformat.bits, &sformat.channels, &sformat.rate);
sprintf(lmsg, "AUDIO: %d Hz, %d ch, xxx, %d kbit\n", sformat.rate, sformat.channels, 0); cs(lmsg);

 fc14_update_info(input_filename);

 fc14state=3; fc14i=1;
 fc14_play();
}
void oscp_fc14_check_input();
void oscp_fc14_parse(char *a);
/////################# check messages
void fc14_check_input() { lf("fc14_check_input"); static char lmsg[4096]; static int rc;
 static fd_set fc14fsa;  static struct timeval fc14tv,*tvp;
 if(fc14state==4) { tvp=NULL; }
 else { fc14tv.tv_sec=0; fc14tv.tv_usec=1; tvp=&fc14tv; }

 FD_ZERO(&fc14fsa); if(parent_said>=0) FD_SET(parent_said, &fc14fsa);

 rc=select(FD_SETSIZE, &fc14fsa, NULL, NULL, tvp); if(rc==0) return;
 else if(rc<0) { sprintf(lmsg, "Info: fc14ci rc<0, %d, %s\n", errno, strerror(errno)); cs(lmsg); }
 else if(rc>0) // check if input is available
 { fdreadline(parent_said, lmsg, 1020); oscp_fc14_parse(lmsg); }
}
/////###################
void fc14_pause_toggle() { lf("fc14_pause_toggle");
      if(fc14state==3) { fc14state=4; cs("===== PAUSE =====\n");
  do { oscp_fc14_check_input(); } while(fc14state==4);
 }
 else if(fc14state==4) { fc14state=3; cs("===== RESUME =====\n"); }
 else cs("Info: pause called but fc14state != (3 || 4)"); }

/////##############
void fc14_checksd() {}
/////##############
void fc14_checkmd() {}

/////###########
void fc14_play() { lf("fc14_play"); dlog(9, "fc14_play");
 if(fc14state!=3) { cs("EOF code: can't play, not inited\n"); return; }
 float ot1=-7.0, ts=((float)fc14_dbuf_size)/44100.0/2.0/2.0;

 static char lmsg[128]; void *buf[1];

 while(fc14state==3 && fc14dec_song_end(fc14c)==0)
 {
  fc14dec_buffer_fill(fc14c, fc14_dbuf, fc14_dbuf_size);

  fc14cur+=ts; if((int)fc14cur!=(int)ot1)
  { ot1=fc14cur; sprintf(lmsg, "A: %.1f 0 0 0 0 0 0 0\n", fc14cur); cs(lmsg); }

  buf[0]=fc14_dbuf; adev_out(buf, fc14_dbuf_size/2/sformat.channels, sformat.channels, sfmt);

  fc14i++; fc14_check_input();
 }

 adev_close();
 fc14_die("EOF code: finished decoding\n");
 lf("fc14_play:finished");
}

/////########################
void oscp_fc14_load(char *fn) { lf("oscp_fc14_load");
 fc14_load(fn); //fc14_play();
 lf("oscp_fc14_load:finished");
}
/////#####################
void oscp_fc14_seek(int a) { lf("oscp_fc14_seek");
 if(a>0) { fc14dec_seek(fc14c, (a+fc14cur)*1000.0); fc14cur+=a; }
}
/////########################
void oscp_fc14_parse(char *a) { lf("oscp_fc14_parse"); char lmsg[4096];
// sprintf(lmsg,"ofparse start"); dlog(9,lmsg);
 int l=strlen(a); if(l<=0) { sprintf(lmsg,"ofparse: l==%d ?", l); dlog(9,lmsg); return; }
 if(strcmp(a,"q")==0 || strcmp(a,"quit")==0) { fc14_end(); exit(0);
 } else if(strncmp(a,"loadfile ",9)==0) { oscp_fc14_load(a+9);
 } else if(strncmp(a,"load ",5)==0) { oscp_fc14_load(a+5);
 } else if(strncmp(a,"pause",5)==0) { fc14_pause_toggle();
 } else if(strncmp(a,"get_property length",19)==0) {
  sprintf(lmsg, "ANS_length=%.1f\n", fc14dur); cs(lmsg);
 } else if(strncmp(a,"get_property time_pos",19)==0 || strncmp(a,"get_time_pos",12)==0) {
//  sprintf(lmsg, "A: %.1f 0 0 0 0 0 0 0 0 0\n", fc14cur); cs(lmsg);
 } else if(strncmp(a,"seek ",5)==0) { int o1=0;
  if(a[5]==*"-") { sscanf(a+6, "%d", &o1); o1*=-1; }
  else if(a[5]==*"+") { sscanf(a+6, "%d", &o1); }
  else { sscanf(a+5, "%d", &o1); }
  oscp_fc14_seek(o1);
 } else if(strncmp(a,"vis ",4)==0) { sscanf(a+4, "%d", &vis);
 } else if(strncmp(a,"lfdump",5)==0) { lfdump();
 } else { sprintf(lmsg, "ofparse unhandled: %s", a); dlog(9, lmsg); }
 lf("oscp_fc14_parse:finished");
}
/////####################### check stdin
void oscp_fc14_check_input() { lf("oscp_fc14_check_input");
 char lmsg[4096]; int i, rc; fd_set fsa;
 struct timeval *tvp=NULL; //, tv; tv.tv_sec=0; tv.tv_usec=10000*99; // i=(int)time(NULL);
 usleep(10000); FD_ZERO(&fsa); if(parent_said>=0) FD_SET(parent_said, &fsa);
 rc=select(FD_SETSIZE, &fsa, NULL, NULL, tvp); // block till input

 if(rc<0) { if(0) { exit(EXIT_FAILURE); } else usleep(10000*99); }
 else if(rc==0) { }
 else for(i=0; i<FD_SETSIZE; ++i) if(FD_ISSET(i, &fsa)) {
  if(i==parent_said) { fdreadline(i, lmsg, 1020); oscp_fc14_parse(lmsg); }
 }
 lf("oscp_fc14_check_input:finished");
}
////################
int oscp_fc14_main() { lf("oscp_fc14_main");
 cs("oscp-slave fc14 1.0.3\n");
 lf("oscp_fc14_main:fc14_init");
 fc14_init();
 lf("oscp_fc14_main:while");
 while(1) { oscp_fc14_check_input(); }
 lf("oscp_fc14_main:fc14_end");
 fc14_end();
 lf("oscp_fc14_main:finished");
 cs("EOF code: exit\n");
// lfdump();
 return 0;
}
